using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;

using NST = Nemerle.Compiler.NamespaceTree;

namespace Oyster.Units.Macros
{
	/// Contains type generation methods
	module TypeGenerator
	{
		/// Contains already initialized environments
		mutable InitedEnvs : list[GlobalEnv] = [];

		/// Generates new unit type
		public GenerateUnitType(env : GlobalEnv, name : string, coef : double, powers : list[int]) : TypeBuilder
		{
			def nameIsEmpty = name == "";

			// Form type name
			def n = if (nameIsEmpty) Macros.NewSymbol("Unit") else Macros.UseSiteSymbol(name);

			// Try to find this type before creating it
			when (env.LookupType([ n.Id ]).HasValue)
			{
				Message.FatalError($"Type $(env.CurrentNamespace).$name already exists")
			};

			def ifaceName = Util.UseSiteExpr(GetUnitInterface(powers).FullName);
			def t = env.Define(<[ decl:
				[NormCoef($(coef : double))]
				public struct $(n : name)/*['a]*/ : $ifaceName //where 'a : $(ifaceName : name)
				{
					_value : double;

					public this(value : double)
					{
						_value = value
					}
					public this(unit : $ifaceName)
					{
						_value = unit.Value * unit.NormCoef / $(coef : double)
					}

					public Value : double
					{
						get { _value }
					}
					public NormCoef : double
					{
						get { $(coef : double) }
					}

					override public ToString() : string
					{
						_value.ToString()
					}

					static public @:>(unit : $(n : name)) : double
					{
						unit._value
					}
					static public @:(value : double) : $(n : name)
					{
						$(n : name)(value)
					}
					// TODO : Type conversion

					static public @++(unit : $(n : name)) : $(n : name)
					{
						$(n : name)(unit.Value + 1)
					}
					static public @--(unit : $(n : name)) : $(n : name)
					{
						$(n : name)(unit.Value - 1)
					}
					static public @+(unit : $(n : name)) : $(n : name)
					{
						unit
					}
					static public @-(unit : $(n : name)) : $(n : name)
					{
						$(n : name)(0 - unit.Value)
					}
				}
			]>);

			// Define all simple arithmetic operations
			def defineSimpleOp(opName, valueType, opFun, isLeft)
			{
				mutable args = [ <[ parameter: value : $valueType ]>, <[ parameter: unit : $(n : name) ]> ];
				mutable opArgs = [ <[ value ]>, <[ unit.Value ]> ];
				when (!isLeft)
				{
					args = args.Reverse();
					opArgs = opArgs.Reverse();
				};
				
				t.Define(<[ decl:
					static public $(opName.name : name)( .. $args ) : $(n : name)
					{
						$(n : name)($(opFun(opArgs.Head, opArgs.Tail.Head)))
					}
				]>)
			};
			def defineOp(opName, opFun)
			{
				[ true, false ].Iter(defineSimpleOp(opName, <[ double ]>, opFun, _))
			};

			defineOp(<[ @+ ]>, Util.AddExpr);
			defineOp(<[ @- ]>, Util.SubExpr);
			defineOp(<[ @* ]>, Util.MulExpr);
			defineOp(<[ @/ ]>, Util.DivExpr);
			
			t.Compile();
			t
		}

		/// Generates or gets unit interface for given powers list
		GetUnitInterface(powers : list[int]) : TypeInfo
		{
			def env = Util.UnitsEnv;

			def ifaceName = $"$(Util.UnitInterfacePrefix)$(Util.GetUnitSuffix(powers))";
			match (env.LookupType([ ifaceName ]))
			{
				| Some(t) => t
				| _ =>
					// Create new interface
					def ifaceName = Macros.UseSiteSymbol(ifaceName);
					def t = env.Define(<[ decl:
						public interface $(ifaceName : name) : IUnit
						{
							// We must declare all members here because of compiler bug
							Value : double { get; }
							NormCoef : double { get; }
						}
					]>);
					t.Compile();
					t
			}
		}
	}
}
