using System;
using Nemerle;
using Nemerle.Collections;
using Nemerle.Compiler;

namespace Oyster.Units.Macros
{
	using CTI = CompileTimeInfo;

	/// Assembly macro-attribute used to define unit types with custom DSL.
	[MacroUsage(MacroPhase.BeforeInheritance, MacroTargets.Assembly)]
	macro UnitsDefinition(defExpr)
	{
		// Parse top-level declaration elements
		match (defExpr)
		{
			| <[ [ .. $unitDefs ] ]> =>
				// Parse every unit definition block
				foreach (<[ $blockName ($nsName) ( .. $units ) ]> in unitDefs)
				{
					// Define some settings basing on block name
					def (addToBasis, addToAliases) =
						match (blockName)
						{
							| <[ basisUnits ]> => (true, true)
							| <[ basisAliases ]> => (false, true)
							| <[ units ]> => (false, false)
							| _ => Message.FatalError($"Expected block name, got $blockName")
						};

					// Extract namespace name
					def nsName =
						match (nsName)
						{
							| <[ _ ]> => ""
							| <[ $(n : name) ]> => n.Id
							| _ => Message.FatalError($"Expected namespace name, got $nsName")
						};

					// Proceed with units. unitsList will contain tuples (name, coef, [powers]) at the end
					mutable unitsList = [];
					foreach (unitDef in units)
					{
						def parseUnit(unit)
						{
							| <[ $(unitName : name) = $aliasDef - $coefDef ]> =>
								def unitName = unitName.Id;

								// Gather powers list
								mutable havingDefinedAlias = true;
								def powers =
									match (aliasDef)
									{
										| <[ _ ]> =>
											havingDefinedAlias = false;
											if (addToBasis) [] else CTI.PowersMap[unitName]
										
										| _ =>
											def parseAlias(alias)
											{
												| <[ ($(n : name) ^ $(k : int)) ]> => CTI.PowersMap[n.Id].Map(_ * k)

												| <[ $e * ($n ^ $k) ]> =>
													List.Map2(parseAlias(e), parseAlias(<[ ($n ^ $k) ]>), _ + _)

												| <[ $e / ($n ^ $k) ]> =>
													List.Map2(parseAlias(e), parseAlias(<[ ($n ^ $k) ]>), _ - _)

												| <[ $(n : name) ]> => parseAlias(<[ ($(n : name) ^ 1) ]>)

												| <[ $e * $(n : name) ]> => parseAlias(<[ $e * ($(n : name) ^ 1) ]>)

												| <[ $e / $(n : name) ]> => parseAlias(<[ $e / ($(n : name) ^ 1) ]>)

												| _ => Message.FatalError($"Expected unit alias, got $aliasDef")
											};
											def powers = parseAlias(aliasDef);

											// Global aliases must be added immediately
											when (addToAliases)
											{
												CTI.PowersMap[unitName] = powers
											};

											powers
									};

								// Parse coefficient
								def coef =
									match (coefDef)
									{
										| <[ _ ]> => if (addToAliases || !havingDefinedAlias) 1.0 else double.NaN
										| _ =>
											def parseCoef(coef)
											{
												| <[ $(d : double) ]> => d
												| <[ $e * $(d : double) ]> => parseCoef(e) * d
												| <[ $e / $(d : double) ]> => parseCoef(e) / d
												| _ => Message.FatalError($"Expected unit normalization coefficient, got $coefDef")
											};
											parseCoef(coefDef)
									};

								(unitName, coef, powers)

							| <[ $(unitName : name) - $coefDef ]> => parseUnit(<[ $(unitName : name) = _ - $coefDef ]>)

							| <[ $(unitName : name) = $aliasDef ]> => parseUnit(<[ $(unitName : name) = $aliasDef - _ ]>)
									
							| <[ $_ - $_ ]> => Message.FatalError($"Expected unit definition, got $unitDef (transformed into $unit)")

							| <[ $e ]> => parseUnit(<[ $e - _ ]>)
						};
						unitsList ::= parseUnit(unitDef)
					};
					unitsList = unitsList.Reverse();

					def names = unitsList.Map(fun(n, _, _) { n });

					// Maybe initialize basis units list
					when (addToBasis)
					{
						// Calculate powers for basis types
						unitsList = unitsList.Map(
							fun(n, coef, _)
							{
								(n, coef, unitsList.Map(fun(n2, _, _) { if (n == n2) 1 else 0 }))
							});

						// Add all powers to hash
						unitsList.Iter(fun(n, _, powers) { CTI.PowersMap[n] = powers });

						// Initialize BasisUnits
						CTI.BasisUnits = names
					};

					// Maybe add units to global aliases list
					when (addToAliases)
					{
						CTI.CommonUnits = CTI.CommonUnits.Append(names)
					};

					// Generate all new types for units in given NS.
					// Maybe also generate additional aliases if all basis units are declared
					mutable basisCoef = [];
					when (!addToAliases && CTI.BasisUnits.ForAll(names.Contains))
					{
						def additionalNames = CTI.CommonUnits.Filter(fun(n) { !names.Contains(n) });
						unitsList = unitsList.Append(additionalNames.Map(fun(n) { (n, double.NaN, CTI.PowersMap[n]) }));

						// Fill local coef hash
						basisCoef = CTI.BasisUnits.Map(fun(n) { unitsList.Find(fun(n2, _, _) { n == n2 }).Value[1] })
					};

					def env = Util.GetUnitsEnv(nsName);

					foreach ((unitName, coef, powers) in unitsList)
					{
						def coef =
							if (double.IsNaN(coef))
							{
								// Try to obtain coefficient automatically
								List.FoldLeft2(basisCoef, powers, 1.0, fun(c, p : int, r) { Math.Pow(c, p) * r })
							}
							else
							{
								coef
							};

						// Generate type for given namespace
						_ = TypeGenerator.GenerateUnitType(env, unitName, coef, powers)
					}
				}
			| _ => Message.FatalError($"Expected comma-separated unit declaration blocks, got $defExpr")
		}
	}
}
